#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2020 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.


"""Save a user to the database"""


__all__ = ["UserSaver"]


import dateparser

from .baseitemsaver import BaseItemSaver
from .models import Hashtag, Tweet, Url, User
from ..config import Config


class UserSaver(BaseItemSaver):
    """Save a user to the database"""

    def _save_user(self, data, session):
        """Save a user to the database"""
        with Config() as config:
            if config["pseudonymise"]:
                data = User._pseudonymise_api_data(data)

        with session.begin():
            user = session.get(User, data["id"])

            if user is None:
                user = User(
                    id=data["id"],
                    username=data["username"],
                    name=data["name"],
                    description=data["description"],
                    location=data["location"],
                    protected=data["protected"],
                    verified=data["verified"]
                )
                session.add(user)

                user.created_at = dateparser.parse(data["created_at"])

                user.pinned_tweet = (
                    session.get(Tweet, int(data["pinned_tweet_id"]))
                    or Tweet(id=int(data["pinned_tweet_id"]))
                )

                user.profile_image_url = (
                    session.query(Url)
                    .filter(Url.url == data["profile_image_url"])
                    .first()
                ) or Url(url=data["profile_image_url"])

                if (
                    "entities" in data
                    and "description" in data["entities"]
                    and "hashtags" in data["entities"]["description"]
                ):
                    for hashtag in data["entities"]["description"]["hashtags"]:
                        hashtag = (
                            session.get(Hashtag, hashtag)
                            or Hashtag(hashtag=hashtag)
                        )
                        user.hashtags.append(hashtag)

                if (
                    "entities" in data
                    and "url" in data["entities"]
                    and "urls" in data["entities"]["url"]
                ):
                    for url in data["entities"]["url"]["urls"]:
                        url = (
                            session.query(Url)
                            .filter(Url.url == url)
                            .first()
                        ) or Url(url=url)
                        user.urls.append(url)

        return user



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2020 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.


"""Save a batch of data as returned by TweetDownloader.batches"""


__all__ = ["ApiResponseSaver"]


import datetime

import dateparser

from .includessaver import IncludesSaver
from .models import SearchTerm
from .tweetsaver import TweetSaver
from ..engine import Session
from ..tweetdownloader import MonthlyQuotaExceededError


class ApiResponseSaver(IncludesSaver, TweetSaver):
    """Save a batch of data as returned by TweetDownloader.batches"""

    def __init__(self, shutdown):
        """Initialise an ApiResponseSaver instance."""
        super().__init__(shutdown)
        self.count = 0

    def save_batch(self, batch, search_term):
        """Save the data in `batch` to the database"""
        earliest_tweet_created_at = datetime.datetime.now(datetime.timezone.utc)

        with Session() as session:
            with session.begin():
                search_term = (
                    session.query(SearchTerm)
                    .filter(SearchTerm.search_term == search_term)
                    .first()
                ) or session.add(SearchTerm(search_term=search_term))

            if "data" in batch:
                if "includes" in batch["data"]:
                    self._save_includes(batch["data"]["includes"], session)
                for tweet in batch["data"]:
                    if self.shutdown.is_set():
                        break
                    self._save_tweet(tweet, session, search_term)
                    self.count += 1
                    earliest_tweet_created_at = min(
                        earliest_tweet_created_at, dateparser.parse(tweet["created_at"])
                    )
            else:
                if "title" in batch and batch["title"] == "UsageCapExceeded":
                    raise MonthlyQuotaExceededError()
        return earliest_tweet_created_at



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2020 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.


"""Save a media_item to the database"""


__all__ = ["MediaItemSaver"]


from .baseitemsaver import BaseItemSaver
from .models import MediaItem, MediaType


class MediaItemSaver(BaseItemSaver):
    """Save a media_item to the database"""

    def _save_media_item(self, data, session):
        """Save a media_item to the database"""
        with session.begin():
            media_item = session.get(MediaItem, data["media_key"])

            if media_item is None:
                media_item = MediaItem(
                    media_key=data["media_key"],
                    url=data["url"]
                )
                media_item.media_type = (
                    session.query(MediaType)
                    .filter(MediaType.media_type == data["type"])
                    .first()
                ) or MediaType(media_type=data["type"])

                session.add(media_item)

        return media_item



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2019 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.

"""A Twitter SearchTerm ORM data model."""


__all__ = ["SearchTerm"]


import dataclasses

import sqlalchemy
import sqlalchemy.orm

from .base import Base


@dataclasses.dataclass
class SearchTerm(Base):
    """A search_term ORM data model."""

    id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
    search_term = sqlalchemy.Column(sqlalchemy.Text, unique=True)
    tweets = sqlalchemy.orm.relationship(
        "Tweet",
        secondary="search_term_tweet_associations",
        back_populates="search_terms"
    )


class SearchTermTweetAssociation(Base):
    """A many-to-many relation table between URLs and tweets."""

    search_term_id = sqlalchemy.Column(
        sqlalchemy.Integer,
        sqlalchemy.ForeignKey("search_terms.id"),
        primary_key=True,
    )
    tweet_id = sqlalchemy.Column(
        sqlalchemy.BigInteger,
        sqlalchemy.ForeignKey("tweets.id"),
        primary_key=True
    )

#    search_term = sqlalchemy.orm.relationship(
#        "SearchTerm",
#        back_populates="tweets"
#    )
#    tweet = sqlalchemy.orm.relationship(
#        "Tweet",
#        back_populates="search_terms"
#    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2020 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.


"""Worker threads wrapping an APIDownloader."""


__all__ = ["TweetDownloaderThread"]


import datetime
import threading
import time

from .timespan import TimeSpan
from .tweetdownloader import (
    MonthlyQuotaExceededError,
    TemporaryApiResponseError,
    TweetDownloader,
)
from .database.apiresponsesaver import ApiResponseSaver


class TweetDownloaderThread(threading.Thread):
    """Wraps a TweetDownloader to run in a separate thread."""

    def __init__(self, api_key_manager, todo_deque, done_queue):
        """
        Intialize an PhotoDownloaderThread.

        Args:
            api_key_manager: instance of an ApiKeyManager
            todo_deque: collections.deque that serves (search_term, TimeSpan)
                        tuples that need to be downloaded
            done_queue: queue.Queue into which to put (search_term, TimeSpan)
                        tuples that have been downloaded

        """
        super().__init__()

        self._api_key_manager = api_key_manager
        self._todo_deque = todo_deque
        self._done_queue = done_queue

        self.shutdown = threading.Event()
        self.api_response_saver = ApiResponseSaver(self.shutdown)

    @property
    def count(self):
        """Count how many tweets we saved to the database"""
        try:
            count = self.api_response_saver.count
        except AttributeError:
            count = -1
        return count

    def run(self):
        """Get TimeSpans off todo_deque and download photos."""
        while not self.shutdown.is_set():
            try:
                search_term, timespan = self._todo_deque.pop()
                self._search_term = search_term
            except IndexError:
                break

            tweet_downloader = TweetDownloader(
                search_term, timespan, self._api_key_manager
            )

            earliest_tweet = timespan.end  # haven’t covered anything yet
            try:
                for batch in tweet_downloader.batches:
                    earliest_tweet = min(
                        earliest_tweet,
                        self.api_response_saver.save_batch(batch, search_term)
                    )

                    if self.shutdown.is_set():
                        timespan.start = earliest_tweet
                        break

            except TemporaryApiResponseError as exception:
                # (includes RateLimitExceededError)
                # report what we managed to download ...
                self._done_queue.put(
                    (search_term, TimeSpan(earliest_tweet, timespan.end))
                )

                # and remember what we haven’t been able to download
                timespan.end = earliest_tweet
                self._todo_deque.append((search_term, timespan))

                # then wait until we’re allowed again
                wait_seconds = (
                    exception.reset_time - datetime.datetime.now(datetime.timezone.utc)
                ).total_seconds()
                for _ in range(wait_seconds):
                    time.sleep(1)
                    if self.shutdown.is_set():
                        break
                else:
                    continue

            except MonthlyQuotaExceededError as exception:
                # TODO: report error properly,
                # for now, re-raise exception to escalte to parent thread
                raise exception from None

            # … report to parent thread how much we worked
            self._done_queue.put((search_term, timespan))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2020 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.


"""Download (all) tweets."""


__all__ = ["BasicTwitterHistoryDownloader"]


import collections
import datetime
import queue
import math
import multiprocessing
import sys
import threading
import time

from .apikeymanager import ApiKeyManager
from .cache import Cache
from .cacheupdaterthread import CacheUpdaterThread
from .config import Config
from .tweetdownloaderthread import TweetDownloaderThread
from .sigtermreceivedexception import SigTermReceivedException
from .timespan import TimeSpan


class BasicTwitterHistoryDownloader:
    """Download (all) tweets."""

    NUM_WORKERS = multiprocessing.cpu_count()

    NUM_MANAGERS = 2  # main thread + cache_updater

    # if output into pipe (e.g. logger, systemd), then
    # print status every 10 minutes, else every 1/10 sec
    # also normal linefeed instead of carriage return for piped output
    STATUS_UPDATE_SEC = 0.1 if sys.stderr.isatty() else 600
    STATUS_UPDATE_LINE_END = "\r" if sys.stderr.isatty() else "\n"

    def __init__(self):
        """Intialise a BasicTwitterHistoryDownloader."""
        self.started = datetime.datetime.now()

        self._todo_deque = collections.deque()
        self._done_queue = queue.Queue()

        self._worker_threads = []
        self._cache_updater_thread = CacheUpdaterThread(self._done_queue)

        with Config() as config:
            self._api_key_manager = ApiKeyManager(
                [config["twitter_oauth2_bearer_token"]],
                300.0 / (15.0 * 60.0)  # 300/15 min,
                # https://developer.twitter.com/en/docs/twitter-api/tweets/
                # search/api-reference/get-tweets-search-all
            )

    def download(self):
        """Download all tweets."""
        with Config() as config:
            for search_term in config["search_terms"]:
                for gap in self.gaps_in_download_history(search_term):
                    self._todo_deque.appendleft((search_term, gap))

        try:
            # start downloaders
            for _ in range(self.NUM_WORKERS):
                worker = TweetDownloaderThread(
                    self._api_key_manager, self._todo_deque, self._done_queue
                )
                worker.start()
                self._worker_threads.append(worker)

            # start cache updater
            self._cache_updater_thread = CacheUpdaterThread(self._done_queue)
            self._cache_updater_thread.start()

            while threading.active_count() > self.NUM_MANAGERS:
                self.report_progress()
                time.sleep(self.STATUS_UPDATE_SEC)

        except (KeyboardInterrupt, SigTermReceivedException):
            self.announce_shutdown()
            for worker in self._worker_threads:
                worker.shutdown.set()

        finally:
            self.summarise_overall_progress()
            for worker in self._worker_threads:
                worker.join()
            self._cache_updater_thread.shutdown.set()
            self._cache_updater_thread.join()

    def report_progress(self):
        """Report current progress."""
        tweet_count, _ = self._statistics
        print(
            (
                "Downloaded metadata for {tweets: 6d} tweets "
                + "using {workers:d} workers, "
                + "{todo:d} time slots to cover"
            ).format(
                tweets=tweet_count,
                workers=(threading.active_count() - self.NUM_MANAGERS),
                todo=len(self._todo_deque),
            ),
            file=sys.stderr,
            end=self.STATUS_UPDATE_LINE_END,
            flush=True
        )

    @classmethod
    def announce_shutdown(cls):
        """Tell the user that we initiated shutdown."""
        print(
            "Cleaning up" + (" " * 69),  # 80 - len("Cleaning up")
            file=sys.stderr,
            end=cls.STATUS_UPDATE_LINE_END,
            flush=True
        )

    def summarise_overall_progress(self):
        """
        Summarise what we have done.

        (Called right before exit)
        """
        tweet_count, _ = self._statistics
        print(
            "Downloaded {tweets:d} tweets ".format(tweets=tweet_count),
            file=sys.stderr,
            flush=True
        )

    @staticmethod
    def gaps_in_download_history(search_term):
        """Find gaps in download history."""
        already_downloaded = BasicTwitterHistoryDownloader.already_downloaded_timespans(
            search_term
        )
        one_day = datetime.timedelta(days=1)  # for comparison

        for i in range(len(already_downloaded) - 1):
            gap = TimeSpan(already_downloaded[i].end, already_downloaded[i + 1].start)
            if gap.duration > one_day:
                divider = math.ceil(gap.duration / one_day)
                for part_of_gap in gap / divider:
                    yield part_of_gap
            else:
                yield gap

    @staticmethod
    def already_downloaded_timespans(search_term):
        """Figure out for which time spans we already have data."""
        with Cache() as cache:
            try:
                timespans = cache[search_term]["already downloaded"]
            except KeyError:
                timespans = []

        # delete existing 0-length time spans
        timespans = [
            timespan
            for timespan in timespans
            if timespan.duration > datetime.timedelta(0)
        ]

        # add 0-length time spans for
        # - first ever Tweet (https://twitter.com/jack/status/20)
        # - now()
        zero = datetime.datetime(2006, 3, 21, 22, 50, 0, tzinfo=datetime.timezone.utc)
        now = datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(
            minutes=1
        )
        timespans += [TimeSpan(zero, zero), TimeSpan(now, now)]

        return sum(timespans)  # sum resolves overlaps

    @property
    def _statistics(self):
        runtime = float((datetime.datetime.now() - self.started).total_seconds())

        tweet_count = sum(
            [
                worker.count
                for worker in self._worker_threads
                if isinstance(worker, TweetDownloaderThread)
            ]
        )
        tweet_rate = tweet_count / runtime

        return (tweet_count, tweet_rate)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2019 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.

"""A Twitter MediaItem ORM data model."""


__all__ = ["MediaItem"]


import dataclasses

import sqlalchemy
import sqlalchemy.orm

from .base import Base


# https://developer.twitter.com/en/docs/twitter-api/data-dictionary/object-model/media


@dataclasses.dataclass
class MediaType(Base):
    """A media type ORM data model."""

    id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
    media_type = sqlalchemy.Column(sqlalchemy.Text, unique=True)
    media_items = sqlalchemy.orm.relationship("MediaItem")


@dataclasses.dataclass
class MediaItem(Base):
    """A Twitter MediaItem ORM data model."""

    media_key = sqlalchemy.Column(sqlalchemy.Text, primary_key=True)
    url = sqlalchemy.Column(sqlalchemy.Text)
    media_type_id = sqlalchemy.Column(
        sqlalchemy.Integer, sqlalchemy.ForeignKey("media_types.id")
    )
    media_type = sqlalchemy.orm.relationship("MediaType", back_populates="media_items")

    tweets = sqlalchemy.orm.relationship(
        "Tweet", secondary="media_item_tweet_associations", back_populates="media"
    )


class MediaItemTweetAssociation(Base):
    """A many-to-many relation table between media items and tweets."""

    media_item_media_key = sqlalchemy.Column(
        sqlalchemy.Text,
        sqlalchemy.ForeignKey("media_items.media_key"),
        primary_key=True,
    )
    tweet_id = sqlalchemy.Column(
        sqlalchemy.BigInteger,
        sqlalchemy.ForeignKey("tweets.id"),
        primary_key=True
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2019 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.

"""A Twitter User ORM data model."""


__all__ = ["User"]


import dataclasses

import sqlalchemy
import sqlalchemy.orm

from .base import Base


# https://developer.twitter.com/en/docs/twitter-api/data-dictionary/object-model/place


class Mention(Base):
    """A many-to-many relation table between users and tweets."""

    tweet_id = sqlalchemy.Column(
        sqlalchemy.BigInteger, sqlalchemy.ForeignKey("tweets.id"), primary_key=True
    )
    user_id = sqlalchemy.Column(
        sqlalchemy.BigInteger, sqlalchemy.ForeignKey("users.id"), primary_key=True
    )


@dataclasses.dataclass
class User(Base):
    """A Twitter User ORM data model."""

    id = sqlalchemy.Column(sqlalchemy.BigInteger, primary_key=True)
    username = sqlalchemy.Column(sqlalchemy.Text, index=True)
    name = sqlalchemy.Column(sqlalchemy.Text)
    description = sqlalchemy.Column(sqlalchemy.Text)
    location = sqlalchemy.Column(sqlalchemy.Text)
    created_at = sqlalchemy.Column(sqlalchemy.DateTime(timezone=True))

    protected = sqlalchemy.Column(sqlalchemy.Boolean)
    verified = sqlalchemy.Column(sqlalchemy.Boolean)

    pinned_tweet_id = sqlalchemy.Column(
        sqlalchemy.BigInteger, sqlalchemy.ForeignKey("tweets.id")
    )
    pinned_tweet = sqlalchemy.orm.relationship("Tweet", foreign_keys=[pinned_tweet_id])

    profile_image_url_id = sqlalchemy.Column(
        sqlalchemy.BigInteger, sqlalchemy.ForeignKey("urls.id")
    )
    profile_image_url = sqlalchemy.orm.relationship("Url")

    hashtags = sqlalchemy.orm.relationship(
        "Hashtag", secondary="hashtag_user_associations", back_populates="users"
    )
    urls = sqlalchemy.orm.relationship(
        "Url", secondary="url_user_associations", back_populates="users"
    )

    tweets = sqlalchemy.orm.relationship(
        "Tweet", back_populates="author", foreign_keys="[Tweet.author_id]"
    )

    mentioned = sqlalchemy.orm.relationship(
        "Tweet", secondary="mentions", back_populates="mentions"
    )

    @classmethod
    def _pseudonymise_api_data(cls, data):
        for identifier in [
            "id",
            "pinned_tweet_id",
        ]:
            try:
                data[identifier] = cls.integer_hash(data[identifier])
            except KeyError:
                pass

        for identifier in [
            "username",
            "name",
            "description",
        ]:
            data[identifier] = None

        return data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2020 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.


"""Save a place to the database"""


__all__ = ["PlaceSaver"]


import shapely.geometry

from .baseitemsaver import BaseItemSaver
from .models import Country, Place, PlaceType


class PlaceSaver(BaseItemSaver):
    """Save a place to the database"""

    def _save_place(self, data, session):
        """Save a place to the database"""
        with session.begin():
            place = session.get(Place, data["id"])
            session.add(place)

            if place is None:
                place = Place(
                    id=data["id"],
                    name=data["name"],
                    full_name=data["fullname"]
                )

                place.place_type = (
                    session.query(PlaceType)
                    .filter(PlaceType.place_type == data["place_type"])
                    .first()
                ) or PlaceType(place_type=data["place_type"])

                if "contained_within" in data:
                    for containing_place in data["contained_within"]:
                        containing_place = (
                            session.get(Place, containing_place)
                            or Place(id=containing_place)
                        )
                        place.contained_within.append(containing_place)

                if "country" in data and "country_code" in data:
                    place.country = (
                        session.get(Country, data["country_code"])
                        or Country(
                            code=data["country_code"],
                            name=data["country"]
                        )
                    )

                if "geometry" in data["geo"]:
                    place.geom = "SRID=4326;" + shapely.geometry.shape(data["geo"]["geometry"]).wkt
                elif "bbox" in data["geo"]:
                    place.geom = "SRID=4326;" + shapely.geometry.box(*data["geo"]["bbox"]).wkt

        return place



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2020 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.


"""
Download (all) tweets.

Overloaded to provide fancy console output
"""


__all__ = ["FancyTwitterHistoryDownloader"]


import threading

import blessed

from .basictwitterhistorydownloader import BasicTwitterHistoryDownloader
from . import __version__ as version


class FancyTwitterHistoryDownloader(BasicTwitterHistoryDownloader):
    """
    Download (all) tweets.

    With fancy console output.
    """

    WELCOME = (
        "{t.bold}{t.blue} ### twitterhistory "
        + "{t.normal}{t.blue}"
        + version
        + "{t.bold} ###"
        + "{t.normal}"
    )

    STATUS = (
        "{t.normal} Downloaded metadata for "
        + "{t.bold}{t.magenta}{tweets: 9d} 💬 tweets "
        + "{t.normal}{t.magenta}{tweet_rate: 11.1f}/s\n"
        + "{t.normal} using                   "
        + "{t.bold}{t.green}{workers: 9d} 💪 workers\n"
        + "{t.normal}{t.bold} TODO:                {todo: 12d} 🚧 time slots"
        + "{t.normal}"
    )
    STATUS_LINES = len(STATUS.splitlines())

    SHUTDOWN_ANNOUNCEMENT = "{t.bold}Cleaning up. 🛑 {t.normal}"

    SUMMARY = (
        "{t.normal}Downloaded  {t.bold}{t.magenta}{tweets: 9d} 💬 tweets "
        + "{t.normal}{t.magenta}{tweet_rate: 11.1f}/s\n"
        + "{t.normal}"
    )

    def __init__(self):
        """Initialise TweetDownloader."""
        super().__init__()

        self.terminal = blessed.Terminal()

        print(self.WELCOME.format(t=self.terminal))

        # scroll down terminal, in case we’re at the bottom
        print(self.STATUS_LINES * "\n", end="")

        self.pos_y, _ = self.terminal.get_location(timeout=5)
        self._tweet_count = 0

    def report_progress(self):
        """Report current progress."""
        tweet_count, tweet_rate = self._statistics

        with self.terminal.location(0, (self.pos_y - self.STATUS_LINES)):
            print(
                self.STATUS.format(
                    t=self.terminal,
                    tweets=tweet_count,
                    tweet_rate=tweet_rate,
                    workers=(threading.active_count() - self.NUM_MANAGERS),
                    todo=len(self._todo_deque),
                )
            )

    def announce_shutdown(self):
        """Tell the user that we initiated shutdown."""
        # clear the status output
        for i in range(self.STATUS_LINES):
            with self.terminal.location(0, (self.pos_y - (i + 1))):
                print(self.terminal.clear_eol)

        with self.terminal.location(0, (self.pos_y - self.STATUS_LINES)):
            print(self.SHUTDOWN_ANNOUNCEMENT.format(t=self.terminal))

    def summarise_overall_progress(self):
        """
        Summarise what we have done.

        (Called right before exit)
        """
        tweet_count, tweet_rate = self._statistics
        with self.terminal.location(0, (self.pos_y - self.STATUS_LINES)):
            print(
                self.SUMMARY.format(
                    t=self.terminal,
                    tweets=tweet_count,
                    tweet_rate=tweet_rate,
                )
            )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2020 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.


"""Download posts and user metadata from the microblogging service Twitter."""


__all__ = ["TwitterHistoryDownloader"]


try:
    import blessed
    from .fancytwitterhistorydownloader import FancyTwitterHistoryDownloader
except ImportError:
    blessed = False

from .basictwitterhistorydownloader import BasicTwitterHistoryDownloader


class TwitterHistoryDownloader:  # pylint: disable=too-few-public-methods
    """Download posts and user metadata from the microblogging service Twitter."""

    __abstract__ = True

    def __new__(cls, *args, **kwargs):
        """Create a new TweetDownloader (dep. on terminal’s capabilities)."""
        if blessed and blessed.Terminal().does_styling:
            _cls = FancyTwitterHistoryDownloader
        else:
            _cls = BasicTwitterHistoryDownloader

        instance = _cls.__new__(_cls, *args, **kwargs)
        instance.__init__(*args, **kwargs)
        return instance



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2019 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.

"""Update the database schema if necessary."""


__all__ = ["DatabaseSchemaUpdater"]


import sys

import sqlalchemy

from ..engine import engine


# for now, schema updates are SQL only and work on PostgreSQL, only.
# GeoAlchemy2 doesn’t really support SQLite, anyway
SCHEMA_UPDATES = {
    # 0 -> 1
    1: """
        BEGIN;

        ALTER TABLE
            places
        ADD COLUMN
            country_code TEXT;

        UPDATE
            places p
        SET
            country_code = c.code
        FROM
            countries c
        WHERE
            p.country_name = c.name;

        ALTER TABLE
            places
        DROP COLUMN
            country_name
                CASCADE;

        ALTER TABLE
            countries
        DROP CONSTRAINT
            countries_pkey,
        DROP CONSTRAINT
            countries_code_key,
        ADD CONSTRAINT
            countries_pkey PRIMARY KEY (code);

        ALTER TABLE
            places
        ADD CONSTRAINT
            places_country_code_fkey
                FOREIGN KEY (country_code)
                REFERENCES countries (code);

        COMMIT;
    """,
    # 1 -> 2
    2: """
        BEGIN;

        ALTER TABLE
            places
        ADD COLUMN
            geom GEOMETRY('GEOMETRYCOLLECTION', 4326);

        COMMIT;
    """,
    # 2 -> 3
    3: """
        BEGIN;

        ALTER TABLE
            search_terms
        DROP CONSTRAINT
            search_terms_pkey
                CASCADE,
        ADD COLUMN
            id SERIAL PRIMARY KEY;

        ALTER TABLE
            search_term_tweet_associations
        ADD COLUMN
            search_term_id INTEGER REFERENCES search_terms (id);

        UPDATE
            search_term_tweet_associations a
        SET
            search_term_id = s.id
        FROM
            search_terms s
        WHERE
            a.search_term_search_term = s.search_term;

        ALTER TABLE
            search_term_tweet_associations
        DROP COLUMN
            search_term_search_term
                CASCADE,
        ADD CONSTRAINT
            search_term_tweet_associations_pkey
                PRIMARY KEY (tweet_id, search_term_id);

        CREATE INDEX ON
            users
        USING
            BTREE(username);

        COMMIT;
    """,
    # 3 -> 4
    4: """
        BEGIN;

        UPDATE
            countries
        SET
            name = ''
        WHERE
            name IS NULL;

        ALTER TABLE
            countries
        ALTER COLUMN
            name
                SET NOT NULL;

        ALTER TABLE
            media_types
        DROP CONSTRAINT
            media_types_pkey
                CASCADE,
        ADD COLUMN
            id SERIAL PRIMARY KEY,
        ADD UNIQUE
            (media_type);

        ALTER TABLE
            media_items
        ADD COLUMN
            media_type_id INTEGER REFERENCES media_types (id);

        UPDATE
            media_items i
        SET
            media_type_id = t.id
        FROM
            media_types t
        WHERE
            i.media_type_media_type = t.media_type;

        ALTER TABLE
            media_items
        DROP COLUMN
            media_type_media_type
                CASCADE;

        ALTER TABLE
            place_types
        DROP CONSTRAINT
            place_types_pkey
                CASCADE,
        ADD COLUMN
            id SERIAL PRIMARY KEY,
        ADD UNIQUE
            (place_type);

        ALTER TABLE
            places
        ADD COLUMN
            place_type_id INTEGER REFERENCES place_types (id);

        UPDATE
            places p
        SET
            place_type_id = t.id
        FROM
            place_types t
        WHERE
            p.place_type_place_type = t.place_type;

        ALTER TABLE
            places
        DROP COLUMN
            place_type_place_type
                CASCADE;

        ALTER TABLE
            tweet_reference_types
        DROP CONSTRAINT
            tweet_reference_types_pkey
                CASCADE,
        ADD COLUMN
            id SERIAL PRIMARY KEY,
        ADD UNIQUE
            (reference_type);

        ALTER TABLE
            tweet_references
        ADD COLUMN
            reference_type_id INTEGER REFERENCES tweet_reference_types (id);

        UPDATE
            tweet_references r
        SET
            reference_type_id = t.id
        FROM
            tweet_reference_types t
        WHERE
            r.reference_type_reference_type = t.reference_type;

        ALTER TABLE
            tweet_references
        DROP COLUMN
            reference_type_reference_type
                CASCADE;

        ALTER TABLE
            urls
        DROP CONSTRAINT
            urls_pkey
                CASCADE,
        ADD COLUMN
            id BIGSERIAL PRIMARY KEY,
        ADD UNIQUE
            (url);

        ALTER TABLE
            url_tweet_associations
        DROP CONSTRAINT
            url_tweet_associations_pkey
                CASCADE,
        ADD COLUMN
            url_id INTEGER REFERENCES urls (id);

        UPDATE
            url_tweet_associations a
        SET
            url_id = u.id
        FROM
            urls u
        WHERE
            a.url_url = u.url;

        ALTER TABLE
            url_tweet_associations
        ADD PRIMARY KEY
            (url_id, tweet_id),
        DROP COLUMN
            url_url
                CASCADE;

        ALTER TABLE
            url_user_associations
        DROP CONSTRAINT
            url_user_associations_pkey
                CASCADE,
        ADD COLUMN
            url_id INTEGER REFERENCES urls (id);

        UPDATE
            url_user_associations a
        SET
            url_id = u.id
        FROM
            urls u
        WHERE
            a.url_url = u.url;

        ALTER TABLE
            url_user_associations
        ADD PRIMARY KEY
            (url_id, user_id),
        DROP COLUMN
            url_url
                CASCADE;

        ALTER TABLE
            users
        ADD COLUMN
            profile_image_url_id BIGINT REFERENCES urls (id);

        UPDATE
            users u1
        SET
            profile_image_url_id = u2.id
        FROM
            urls u2
        WHERE
            u1.profile_image_url_id IS NOT NULL
        AND
            u1.profile_image_url_url = u2.url;

        ALTER TABLE
            users
        DROP COLUMN
            profile_image_url_url
                CASCADE;

        COMMIT;
    """,
    # 4 -> 5
    5: """
        BEGIN;

        ALTER TABLE
            tweet_references
        DROP CONSTRAINT
            tweet_references_pkey,
        ADD PRIMARY KEY (
            referencing_tweet_id,
            referenced_tweet_id,
            reference_type_id
        );

        COMMIT;
    """
}


# Try to create database table for schema version
with engine.begin() as connection:
    connection.execute(
        """
            CREATE TABLE IF NOT EXISTS
                schema_versions
                (
                    update TIMESTAMP WITH TIME ZONE DEFAULT NOW(),
                    version INTEGER PRIMARY KEY
                );
        """
    )


class DatabaseSchemaUpdater:
    """Update the database schema if necessary."""

    LATEST = "LATEST"  # “magic”, see def set_schema_version

    @property
    def installed_version(self):
        """Current version."""
        with engine.connect() as connection:
            installed_version = (
                connection.execute(
                    sqlalchemy.text("""
                        SELECT
                            COALESCE(
                                MAX(version),
                                0
                            ) AS version
                        FROM
                            schema_versions;
                    """)
                ).scalar_one_or_none()
            )
        return installed_version

    def update_to_latest(self):
        """Update to the latest schema version."""
        installed_version = self.installed_version
        while installed_version < max(SCHEMA_UPDATES.keys()):
            print(
                "Updating database schema (db version {:d}->{:d})".format(
                    installed_version,
                    installed_version + 1
                ),
                file=sys.stderr
            )
            with engine.begin() as connection:
                next_version = self.installed_version + 1
                connection.execute(sqlalchemy.text(SCHEMA_UPDATES[next_version]))
                self.set_schema_version(next_version)
            installed_version = self.installed_version

    @classmethod
    def set_schema_version(cls, version):
        """Set the schema version (without running update scripts)."""
        if version == cls.LATEST:
            version = max(SCHEMA_UPDATES.keys())
        with engine.begin() as connection:
            connection.execute(
                sqlalchemy.text("""
                    INSERT INTO
                        schema_versions (version)
                    VALUES (
                        :version
                    );
                """),
                {
                    "version": version
                }
            )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2019 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.

"""A Twitter Hashtag ORM data model."""


__all__ = ["Hashtag"]


import dataclasses

import sqlalchemy
import sqlalchemy.orm

from .base import Base


@dataclasses.dataclass
class Hashtag(Base):
    """A hashtag ORM data model."""

    hashtag = sqlalchemy.Column(sqlalchemy.Text, primary_key=True)
    tweets = sqlalchemy.orm.relationship(
        "Tweet", secondary="hashtag_tweet_associations", back_populates="hashtags"
    )
    users = sqlalchemy.orm.relationship(
        "User", secondary="hashtag_user_associations", back_populates="hashtags"
    )


class HashtagTweetAssociation(Base):
    """A many-to-many relation table between hashtags and tweets."""

    hashtag_hashtag = sqlalchemy.Column(
        sqlalchemy.Text, sqlalchemy.ForeignKey("hashtags.hashtag"), primary_key=True
    )
    # hashtag = sqlalchemy.orm.relationship("Hashtag", back_populates="tweets")
    tweet_id = sqlalchemy.Column(
        sqlalchemy.BigInteger, sqlalchemy.ForeignKey("tweets.id"), primary_key=True
    )
    # tweet = sqlalchemy.orm.relationship("Tweet", back_populates="hashtags")


class HashtagUserAssociation(Base):
    """A many-to-many relation table between hashtags and users."""

    hashtag_hashtag = sqlalchemy.Column(
        sqlalchemy.Text, sqlalchemy.ForeignKey("hashtags.hashtag"), primary_key=True
    )
    # hashtag = sqlalchemy.orm.relationship("Hashtag", back_populates="users")
    user_id = sqlalchemy.Column(
        sqlalchemy.BigInteger, sqlalchemy.ForeignKey("users.id"), primary_key=True
    )
    # user = sqlalchemy.orm.relationship("User", back_populates="hashtags")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2019 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.

"""A Twitter Place ORM data model."""


__all__ = ["Place"]


import dataclasses

import geoalchemy2
import sqlalchemy
import sqlalchemy.orm

from .base import Base


# https://developer.twitter.com/en/docs/twitter-api/data-dictionary/object-model/place


@dataclasses.dataclass
class PlaceType(Base):
    """A place type ORM data model."""

    id = sqlalchemy.Column(sqlalchemy.BigInteger, primary_key=True)
    place_type = sqlalchemy.Column(sqlalchemy.Text, unique=True)
    places = sqlalchemy.orm.relationship("Place")


@dataclasses.dataclass
class Country(Base):
    """A country ORM data model."""

    __tablename__ = "countries"  # override "countrys"

    name = sqlalchemy.Column(sqlalchemy.Text)
    code = sqlalchemy.Column(sqlalchemy.Text, primary_key=True)
    places = sqlalchemy.orm.relationship("Place")


@dataclasses.dataclass
class Place(Base):
    """A Twitter Place ORM data model."""

    id = sqlalchemy.Column(sqlalchemy.Text, primary_key=True)

    contained_within = sqlalchemy.orm.relationship(
        "Place", back_populates="contains", remote_side=[id]
    )
    contains = sqlalchemy.orm.relationship("Place", back_populates="contained_within")
    container_id = sqlalchemy.Column(
        sqlalchemy.Text, sqlalchemy.ForeignKey("places.id")
    )

    country_code = sqlalchemy.Column(
        sqlalchemy.Text, sqlalchemy.ForeignKey("countries.code")
    )
    country = sqlalchemy.orm.relationship("Country", back_populates="places")

    geom = sqlalchemy.Column(geoalchemy2.Geometry(srid=4326))

    name = sqlalchemy.Column(sqlalchemy.Text)
    full_name = sqlalchemy.Column(sqlalchemy.Text)

    place_type_id = sqlalchemy.Column(
        sqlalchemy.BigInteger, sqlalchemy.ForeignKey("place_types.id")
    )
    place_type = sqlalchemy.orm.relationship("PlaceType", back_populates="places")

    tweets = sqlalchemy.orm.relationship("Tweet")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2019 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.

"""A Tweet ORM data model."""


__all__ = ["Tweet"]


import dataclasses

import geoalchemy2
import sqlalchemy
import sqlalchemy.orm

from .base import Base


# https://developer.twitter.com/en/docs/twitter-api/data-dictionary/object-model/tweet


@dataclasses.dataclass
class TweetReferenceType(Base):
    """A media type ORM data model."""
    id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
    reference_type = sqlalchemy.Column(sqlalchemy.Text, unique=True)


@dataclasses.dataclass
class TweetReference(Base):  # ‘Association Object’ in sqlalchemy terminology
    """An ORM model to represent the relationship between two tweets."""

    referenced_tweet_id = sqlalchemy.Column(
        sqlalchemy.BigInteger,
        sqlalchemy.ForeignKey("tweets.id"),
        primary_key=True
    )
    referenced_tweet = sqlalchemy.orm.relationship(
        "Tweet",
        foreign_keys=[referenced_tweet_id],
        back_populates="referencing_tweets"
    )

    referencing_tweet_id = sqlalchemy.Column(
        sqlalchemy.BigInteger,
        sqlalchemy.ForeignKey("tweets.id"),
        primary_key=True
    )
    referencing_tweet = sqlalchemy.orm.relationship(
        "Tweet",
        foreign_keys=[referencing_tweet_id],
        back_populates="referenced_tweets"
    )

    reference_type_id = sqlalchemy.Column(
        sqlalchemy.Integer,
        sqlalchemy.ForeignKey("tweet_reference_types.id"),
        primary_key=True
    )
    reference_type = sqlalchemy.orm.relationship("TweetReferenceType")


@dataclasses.dataclass
class Tweet(Base):
    """A Tweet ORM data model."""

    id = sqlalchemy.Column(sqlalchemy.BigInteger, primary_key=True)
    created_at = sqlalchemy.Column(sqlalchemy.DateTime(timezone=True))
    possibly_sensitive = sqlalchemy.Column(sqlalchemy.Boolean)
    text = sqlalchemy.Column(sqlalchemy.Text)

    geom = sqlalchemy.Column(geoalchemy2.Geometry("POINT", 4326))

    author_id = sqlalchemy.Column(
        sqlalchemy.BigInteger, sqlalchemy.ForeignKey("users.id")
    )
    author = sqlalchemy.orm.relationship(
        "User", back_populates="tweets", foreign_keys=[author_id]
    )

    in_reply_to_user_id = sqlalchemy.Column(
        sqlalchemy.BigInteger, sqlalchemy.ForeignKey("users.id")
    )
    in_reply_to = sqlalchemy.orm.relationship(
        "User", foreign_keys=[in_reply_to_user_id]
    )

    conversation_id = sqlalchemy.Column(
        sqlalchemy.BigInteger, sqlalchemy.ForeignKey("conversations.id")
    )
    conversation = sqlalchemy.orm.relationship("Conversation", back_populates="tweets")

    language_language = sqlalchemy.Column(
        sqlalchemy.Text, sqlalchemy.ForeignKey("languages.language")
    )
    language = sqlalchemy.orm.relationship("Language", back_populates="tweets")

    place_id = sqlalchemy.Column(sqlalchemy.Text, sqlalchemy.ForeignKey("places.id"))
    place = sqlalchemy.orm.relationship("Place", back_populates="tweets")

    referenced_tweets = sqlalchemy.orm.relationship(
        "TweetReference",
        primaryjoin="(Tweet.id==TweetReference.referencing_tweet_id)",
        back_populates="referencing_tweet",
    )
    referencing_tweets = sqlalchemy.orm.relationship(
        "TweetReference",
        primaryjoin="(Tweet.id==TweetReference.referenced_tweet_id)",
        back_populates="referenced_tweet",
    )

    hashtags = sqlalchemy.orm.relationship(
        "Hashtag", secondary="hashtag_tweet_associations", back_populates="tweets"
    )
    media = sqlalchemy.orm.relationship(
        "MediaItem", secondary="media_item_tweet_associations", back_populates="tweets"
    )
    mentions = sqlalchemy.orm.relationship(
        "User", secondary="mentions", back_populates="mentioned"
    )
    urls = sqlalchemy.orm.relationship(
        "Url", secondary="url_tweet_associations", back_populates="tweets"
    )

    search_terms = sqlalchemy.orm.relationship(
        "SearchTerm",
        secondary="search_term_tweet_associations",
        back_populates="tweets"
    )

    @classmethod
    def _pseudonymise_api_data(cls, data):
        for identifier in [
            "id",
            "author_id",
            "in_reply_to_user_id",
            "conservation_id",
        ]:
            try:
                data[identifier] = cls.integer_hash(data[identifier])
            except KeyError:
                pass

        if "referenced_tweets" in data:
            for i in range(len(data["referenced_tweets"])):
                data["referenced_tweets"][i]["id"] = cls.integer_hash(
                    data["referenced_tweets"][i]["id"]
                )

        for identifier in []:
            data[identifier] = None

        return data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2019 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.

"""ORM models for Twitter entities."""


__all__ = [
    "Conversation",
    "Country",
    "Hashtag",
    "Language",
    "MediaItem",
    "MediaType",
    "Place",
    "PlaceType",
    "SearchTerm",
    "Tweet",
    "TweetReference",
    "TweetReferenceType",
    "Url",
    "User",
]


import sqlalchemy

from .conversation import Conversation
from .hashtag import Hashtag
from .language import Language
from .mediaitem import MediaItem, MediaType
from .place import Country, Place, PlaceType
from ..databaseschemaupdater import DatabaseSchemaUpdater
from .searchterm import SearchTerm
from .tweet import Tweet, TweetReference, TweetReferenceType
from .url import Url
from .user import User

from ...engine import engine
from .base import Base


if sqlalchemy.inspect(engine).has_table(Tweet.__table__):  # DATA_EXISTS
    DatabaseSchemaUpdater().update_to_latest()
else:
    Base.metadata.create_all(engine)
    DatabaseSchemaUpdater().set_schema_version(DatabaseSchemaUpdater.LATEST)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2020 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.


"""Save the `includes` contained in a TweetDownloader.batch"""


__all__ = ["IncludesSaver"]


from .mediaitemsaver import MediaItemSaver
from .placesaver import PlaceSaver
from .tweetsaver import TweetSaver
from .usersaver import UserSaver


class IncludesSaver(MediaItemSaver, PlaceSaver, TweetSaver, UserSaver):
    """Save the `includes` contained in a TweetDownloader.batch"""
    def _save_includes(self, includes, session):
        if "media" in includes:
            for media_item in includes["media"]:
                if self.shutdown.is_set():
                    break
                self._save_media_item(media_item, session)
        if "places" in includes:
            for place in includes["places"]:
                if self.shutdown.is_set():
                    break
                self._save_place(place, session)
        if "tweets" in includes:
            for tweet in includes["tweets"]:
                if self.shutdown.is_set():
                    break
                self._save_tweet(tweet, session)
        if "users" in includes:
            for user in includes["users"]:
                if self.shutdown.is_set():
                    break
                self._save_user(user, session)

        # note to self:
        #   to avoid redundant code here,
        #   should we have an iterator method that yields
        #   the next thing and checks in between?



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2020 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.


"""Save a tweet to the database"""


__all__ = ["TweetSaver"]


import dateparser
import shapely.geometry

from .baseitemsaver import BaseItemSaver
from .models import (
    Conversation,
    Hashtag,
    Language,
    MediaItem,
    Place,
    Tweet,
    TweetReference,
    TweetReferenceType,
    Url,
    User
)
from ..config import Config


class TweetSaver(BaseItemSaver):
    """Save a tweet to the database"""

    def _save_tweet(self, data, session, search_term=None):
        """Save a tweet to the database"""
        with Config() as config:
            if config["pseudonymise"]:
                data = Tweet._pseudonymise_api_data(data)

        with session.begin():
            tweet = session.get(Tweet, data["id"])

            if tweet is None:
                tweet = Tweet(
                    id=data["id"],
                    created_at=dateparser.parse(data["created_at"]),
                    possibly_sensitive=data["possibly_sensitive"],
                    text=data["text"]
                )
                session.add(tweet)

                tweet.author = (
                    session.get(User, data["author_id"])
                    or User(id=data["author_id"])
                    # merging, because we might otherwise run into
                    # a conflict further down in cause the author
                    # replied to themselves
                )

                if (
                    "geo" in data
                    and "coordinates" in data["geo"]
                    and "coordinates" in data["geo"]["coordinates"]
                ):
                    tweet.geom = "SRID=4326;" + shapely.geometry.Point(
                        *data["geo"]["coordinates"]["coordinates"]
                    ).wkt

                if "geo" in data and "place_id" in data["geo"]:
                    tweet.place = (
                        session.get(Place, data["geo"]["place_id"])
                        or Place(id=data["geo"]["place_id"])
                    )

                if "in_reply_to_user_id" in data:
                    if data["in_reply_to_user_id"] == data["author_id"]:
                        tweet.in_reply_to = tweet.author
                    else:
                        tweet.in_reply_to = (
                            session.get(User, data["in_reply_to_user_id"])
                            or User(id=data["in_reply_to_user_id"])
                        )

                if "conversation_id" in data:
                    tweet.conversation = (
                        session.get(Conversation, data["conversation_id"])
                        or Conversation(id=data["conversation_id"])
                    )

                tweet.language = (
                    session.get(Language, data["lang"])
                    or Language(language=data["lang"])
                )

                if "attachments" in data and "media_keys" in data["attachments"]:
                    tweet.media = []
                    for media_key in data["attachments"]["media_keys"]:
                        media_item = (
                            session.get(MediaItem, media_key)
                            or MediaItem(media_key=media_key)
                        )
                        if media_item not in tweet.media:
                            tweet.media.append(media_item)

                if "entities" in data and "hashtags" in data["entities"]:
                    tweet.hashtags = []
                    for hashtag in data["entities"]["hashtags"]:
                        hashtag = (
                            session.get(Hashtag, hashtag["tag"])
                            or Hashtag(hashtag=hashtag["tag"])
                        )
                        if hashtag not in tweet.hashtags:
                            tweet.hashtags.append(hashtag)

                if "entities" in data and "urls" in data["entities"]:
                    tweet.urls = []
                    for url in data["entities"]["urls"]:
                        url = (
                            session.query(Url)
                            .filter(Url.url == url["url"])
                            .first()
                        ) or Url(url=url["url"])
                        if url not in tweet.urls:
                            tweet.urls.append(url)

                if "entities" in data and "mentions" in data["entities"]:
                    tweet.mentions = []
                    for mention in data["entities"]["mentions"]:
                        # if everything went in order, then we should already
                        # have saved the user (mention is by username)
                        mentioned_user = (
                            session.query(User)
                            .filter(User.username == mention["username"])
                            .first()
                        )
                        if (
                                mentioned_user is not None  # TODO!
                                and mentioned_user not in tweet.mentions
                        ):
                            tweet.mentions.append(mentioned_user)

                if "referenced_tweets" in data:
                    for referenced_tweet in data["referenced_tweets"]:
                        tweet_reference = TweetReference()
                        tweet_reference.reference_type = (
                            session.query(TweetReferenceType)
                            .filter(TweetReferenceType.reference_type == referenced_tweet["type"])
                            .first()
                        ) or (
                            TweetReferenceType(
                                reference_type=referenced_tweet["type"]
                            )
                        )
                        tweet_reference.referenced_tweet = (
                            session.get(Tweet, referenced_tweet["id"])
                            or Tweet(id=referenced_tweet["id"])
                        )
                        tweet.referenced_tweets.append(tweet_reference)

                if search_term is not None:
                    tweet.search_terms = [search_term]

        return tweet



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2020 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.


"""Download all tweets in a time span relating to a search query."""


__all__ = ["TweetDownloader"]


import datetime
import json

import requests


class ApiResponseError(BaseException):
    """Raised when API returns bogus data."""


class NoAcademicTwitterAccount(ApiResponseError):
    """Raised when API key not enrolled in academic API programme."""


class MonthlyQuotaExceededError(ApiResponseError):
    """Raised when API blocks because montly quota is used up."""


class TemporaryApiResponseError(ApiResponseError):
    """Raised when API has an error and waiting a bit should help"""

    def __init__(self, *args, time_until_reset=-1, **kwargs):
        super().__init__(*args, **kwargs)
        self.reset_time = (
            datetime.datetime.now(datetime.timezone.utc)
            + datetime.timedelta(seconds=time_until_reset)
        )

    def __str__(self):
        return "{:s}(reset_time={:%Y-%m-%d %H:%M:%S})".format(
            self.__class__.__name__, self.reset_time
        )

    def __repr__(self):
        return str(self)


class RateLimitExceededError(TemporaryApiResponseError):
    """Raised when API blocks because rate limit is reached."""


class TweetDownloader:  # pylint: disable=too-few-public-methods
    """Download all tweets in a time span relating to a search query."""

    API_ENDPOINT_URL = "https://api.twitter.com/2/tweets/search/all"

    def __init__(self, search_term, timespan, api_key_manager):
        """Intialize an TweetDownloader."""
        self._search_term = search_term
        self._timespan = timespan
        self._api_key_manager = api_key_manager

    @property
    def batches(self):
        """Iterate over downloaded batches of tweets + aux data."""
        query = {
            "query": self._search_term,
            "expansions": ",".join(
                [
                    #  "attachments.poll_ids",
                    "attachments.media_keys",
                    "author_id",
                    "entities.mentions.username",
                    "geo.place_id",
                    "in_reply_to_user_id",
                    "referenced_tweets.id",
                    "referenced_tweets.id.author_id",
                ]
            ),
            "media.fields": ",".join(
                [
                    #  "duration_ms",
                    #  "height",
                    #  "media_key",
                    #  "preview_image_url",
                    "type",
                    "url",
                    #  "width",
                    #  "public_metrics",
                    #  "non_public_metrics",
                    #  "organic_metrics",
                    #  "promoted_metrics"
                ]
            ),
            "place.fields": ",".join(
                [
                    #  "contained_within",
                    "country",
                    "country_code",
                    "full_name",
                    "geo",
                    "id",
                    "name",
                    "place_type",
                ]
            ),
            "tweet.fields": ",".join(
                [
                    "attachments",
                    "author_id",
                    # "context_annotations",
                    "conversation_id",
                    "created_at",
                    "entities",
                    "geo",
                    "id",
                    "in_reply_to_user_id",
                    "lang",
                    # "non_public_metrics",
                    # "public_metrics",
                    # "organic_metrics",
                    # "promoted_metrics",
                    "possibly_sensitive",
                    "referenced_tweets",
                    "reply_settings",
                    "source",
                    "text",
                    "withheld",
                ]
            ),
            "user.fields": ",".join(
                [
                    "created_at",
                    "description",
                    "entities",
                    "location",
                    "name",
                    "pinned_tweet_id",
                    "profile_image_url",
                    "protected",
                    # "public_metrics",
                    "url",
                    "username",
                    "verified",
                    "withheld",
                ]
            ),
            "start_time": self._timespan.start.isoformat(),
            "end_time": self._timespan.end.isoformat(),
            "max_results": 500,
        }
        next_token = None

        while True:
            if next_token is not None:
                query["next_token"] = next_token

            with self._api_key_manager.get_api_key() as api_key:
                headers = {"Authorization": "Bearer {:s}".format(api_key)}

                with requests.get(
                    self.API_ENDPOINT_URL, headers=headers, params=query
                ) as response:
                    try:
                        # some strings in Twitter API responses can apparently
                        # contain NUL characters that cause hickups in PostgreSQL
                        results = json.loads(
                            response.text
                            .replace("\\u0000", "\\u2400")
                        )
                    except json.decoder.JSONDecodeError as exception:
                        raise ApiResponseError(response.text) from exception

                    if not response.ok:
                        try:
                            if results["error"]["code"] == 88:
                                raise RateLimitExceededError(
                                    response.headers["x-rate-limit-reset"]
                                )
                        except KeyError:
                            pass

                        try:
                            if results["status"] == 429:
                                # results["title"]: "Too Many Requests"
                                # wait 2 min
                                raise TemporaryApiResponseError(
                                    time_until_reset=120
                                )
                        except KeyError:
                            pass

                        try:
                            if results["title"] == "UsageCapExceeded":
                                raise MonthlyQuotaExceededError()
                        except KeyError:
                            pass

                        try:
                            if results["reason"] == "client-not-enrolled":
                                raise NoAcademicTwitterAccount(results["detail"])
                        except KeyError:
                            pass

                        # other, not anticipated error:
                        raise ApiResponseError((results, response.headers))

            yield results

            try:
                next_token = results["meta"]["next_token"]
            except KeyError:
                break



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2019 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.

"""A common sqlalchemy declarative_base() to share between models."""


__all__ = ["Base"]


import json
import re

import sqlalchemy.ext.declarative
import sqlalchemy.orm

from ...config import Config


CAMEL_CASE_TO_SNAKE_CASE_RE = re.compile(
    "((?<=[a-z0-9])[A-Z]|(?!^)(?<!_)[A-Z](?=[a-z]))"
)


def camel_case_to_snake_case(camel_case):
    """Convert a `camelCase` string to `snake_case`."""
    snake_case = CAMEL_CASE_TO_SNAKE_CASE_RE.sub(r"_\1", camel_case).lower()
    return snake_case


class Base:
    """Template for sqlalchemy declarative_base() to add shared functionality."""

    def __str__(self):
        """Return a str representation."""
        primary_keys = {}
        for pk in self.__mapper__.primary_key:
            try:
                primary_keys[pk.name] = getattr(self, pk.name)
            except AttributeError:  # (not yet set)
                pass
        return "{}({})".format(
            self.__class__.__name__,
            json.dumps(primary_keys)
        )

    @staticmethod
    def pseudonymise_identifiers():
        """Pseudonymise identifiers?."""
        pseudonymise_identifiers = True
        try:
            with Config() as config:
                if config["pseudonymise"] is False:
                    pseudonymise_identifiers = False
        except KeyError:
            pass
        return pseudonymise_identifiers

    @sqlalchemy.orm.declared_attr
    def __tablename__(cls):
        """Return a table name derived from the class name."""
        snake_case = camel_case_to_snake_case(cls.__name__)
        return "{:s}s".format(snake_case)


Base = sqlalchemy.ext.declarative.declarative_base(cls=Base)
